package meerkat.voting.output;

import com.google.protobuf.BoolValue;
import meerkat.protobuf.PollingStation.ScannedData;
import meerkat.protobuf.Voting.SignedEncryptedBallot;
import meerkat.rest.Constants;
import meerkat.rest.ProtobufMessageBodyReader;
import meerkat.rest.ProtobufMessageBodyWriter;
import meerkat.voting.output.outputcommands.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.ws.rs.client.Client;
import javax.ws.rs.client.ClientBuilder;
import javax.ws.rs.client.Entity;
import javax.ws.rs.client.WebTarget;
import javax.ws.rs.core.Response;
import java.io.IOException;

import static meerkat.pollingstation.PollingStationConstants.POLLING_STATION_WEB_SCANNER_SCAN_PATH;

/**
 * A ballot output device for the network. It simply sends details over the wire
 */
public class NetworkVirtualPrinter extends AsyncRunnableOutputDevice {

    private static final Logger logger = LoggerFactory.getLogger(NetworkVirtualPrinter.class);
    private SignedEncryptedBallot signedEncryptedBallot;
    private final WebTarget successfulPrintTarget;

    public NetworkVirtualPrinter(String address) {
        super();
        logger.info("A NetworkVirtualPrinter is constructed");
        Client client = ClientBuilder.newClient();
        client.register(ProtobufMessageBodyReader.class);
        client.register(ProtobufMessageBodyWriter.class);
        successfulPrintTarget = client.target(address).path(POLLING_STATION_WEB_SCANNER_SCAN_PATH);
    }


    /**
     * The NetworkVirtualPrinter actually does nothing for committing.
     * It simply keeps the ballot details for later.
     * When the voter chooses to Cast the ballot, these details are sent over the wire.
     * @param command a CommitOutputCommand with the signed encryption of the ballot
     */
    public void doCommitToBallot(CommitOutputCommand command) {
        logger.debug("entered method doCommitToBallot");
        signedEncryptedBallot = command.getSignedEncryptedBallot();
        command.getCallback().onSuccess(null);
    }


    /**
     * The NetworkVirtualPrinter actually does nothing for auditing.
     */
    public void doAudit(AuditOutputCommand command) {
        logger.debug("entered method doAudit");
        signedEncryptedBallot = null;
        command.getCallback().onSuccess(null);
    }


    /**
     * This is where the magic happens. The signed encrypted ballot is transmitted over the wire
     * @param command
     */
    public void doCastBallot(CastOutputCommand command) {
        logger.debug("entered method doCastBallot");
        ScannedData scannedData = ScannedData.newBuilder()
                .setChannel(null)  //TODO: fill the details for the channel to be able to send here
                .setSignedEncryptedBallot(this.signedEncryptedBallot)
                .build();

        Response response = successfulPrintTarget.request(Constants.MEDIATYPE_PROTOBUF).post(Entity.entity(scannedData, Constants.MEDIATYPE_PROTOBUF));
        BoolValue b = response.readEntity(BoolValue.class);
        response.close();

        if (b.getValue()) {
            command.getCallback().onSuccess(null);
        }
        else {
            command.getCallback().onFailure(new IOException());
        }
    }


    /**
     * The NetworkVirtualPrinter actually does nothing for canceling.
     */
    public void doCancel(CancelOutputCommand command) {
        logger.debug("entered method doCancel");
        signedEncryptedBallot = null;
        command.getCallback().onSuccess(null);
    }

}
